#! /usr/bin/python2
#
# Python version of jackaudio bash script dated 20140605
# Adds a listening thread for remote control
# jackaudio - enable jack to forward audio input to output from boot time
#
# author "Leonardo Gabrielli <l.gabrielli@univpm.it>
# defaults are intended for the BeagleBoard xM



import socket
import subprocess
import time
import threading
import sys, os
import psutil # WARNING: psutil-1.2 differs from version 2.x, as of now wemust-OS 0.2 runs version 1.2.1

#######################
##### VARS - DEFS #####
#######################
import wemust
from wemust import USE_NETDBG, SERVICETYPE, UUID_LEN, NETDBG_MCAST_GROUP, NETDBG_BASE_PORT, NETDBG_PRINT_PORT, NETDBG_JACKD_PORT, NETDBG_JTSER_PORT, NETDBG_JTCLI_PORT, NETDBG_CTRL_PORT

# JACK + JACKTRIP PARAMS
PEER_ADDRESS = wemust.getIPfromWEMUSTRC()

# IMPORT PARAMETERS FROM WEMUSTRC
SyP = wemust.SAByParams()
SyP.setFromFile(wemust.DEFAULT_WEMUSTRC_PATH)
SAMPLERATE = SyP.getSampRateOnly()
PERIOD = SyP.getPeriodOnly()
NCHANNELS = 4 # to get from wemustrc
JACKARGS = ('-P62 -t2000 -dalsa -dhw:0 -r' + str(SAMPLERATE) + ' -p' + str(PERIOD) + ' -n2 -s -i' + str(NCHANNELS) + ' ' + '-o' + str(NCHANNELS))
JTBIN = 'jacktrip' # path to executable

threads = [] # ToDo handle thread list, for proper shutdown

# extract jacktrip port offset from hostname (debian-BBx format)
JT_OFFS = ''.join(ch for ch in socket.gethostname() if ch.isdigit())
if JT_OFFS == '':
    JT_OFFS = 0



global jackd_dbgs, jt_cli_dbgs, jt_srv_dbgs#, avahi_pub_dbgs
jackd_dbgs = wemust.dbgUDPSockTx(NETDBG_MCAST_GROUP,NETDBG_JACKD_PORT)
jt_cli_dbgs = wemust.dbgUDPSockTx(NETDBG_MCAST_GROUP,NETDBG_JTCLI_PORT)
jt_srv_dbgs = wemust.dbgUDPSockTx(NETDBG_MCAST_GROUP,NETDBG_JTSER_PORT)






######################
# CPU freq functions #
######################
def tune_max():
    # enable max CPU freq
    subprocess.Popen('echo -n performance | tee /sys/devices/system/cpu/cpu0/cpufreq/scaling_governor'.split())

#
def tune_min():
    # enable max CPU freq
    subprocess.Popen('echo -n powersave | tee /sys/devices/system/cpu/cpu0/cpufreq/scaling_governor'.split())


#
def tune_avg():
    # enable max CPU freq
    subprocess.Popen('echo -n ondemand | tee /sys/devices/system/cpu/cpu0/cpufreq/scaling_governor'.split())


######################
#     JACK funcs     #
######################
def jack_default():
    jack_proc = subprocess.Popen(('sudo -u debian jackd' + ' ' + JACKARGS).split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT, bufsize=-1)
    jackd_dt = threading.Thread(target=wemust.dispatcherThread, args=(jack_proc.stdout, jackd_dbgs))
    jackd_dt.daemon = True # thread dies with the program
    jackd_dt.start()
    time.sleep(5)

#
def conn_loopback():
    subprocess.call('sudo -u debian jack_connect system:capture_1 system:playback_1'.split())
    subprocess.call(('sudo -u debian jack_connect system:capture_' + str(NCHANNELS) + ' ' + 'system:playback_' + str(NCHANNELS)).split())



#
def disc_loopback():
    subprocess.call('sudo -u debian jack_disconnect system:capture_1 system:playback_1'.split())
    subprocess.call(('sudo -u debian jack_disconnect system:capture_' + str(NCHANNELS) + ' ' + 'system:playback_' + str(NCHANNELS)).split())


# test tone
def test_tone():
    tone_proc = subprocess.Popen('sudo -u debian jack_tone -o0.2 -a0.2 -f1174.66'.split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT, bufsize=-1)
    tone_dt = threading.Thread(target=wemust.dispatcherThread, args=(tone_proc.stdout, jackd_dbgs))
    tone_dt.daemon = True
    tone_dt.start()
    time.sleep(1)

    subprocess.call('sudo -u debian jack_connect jack_tone:out JackTrip:send_1'.split())
    subprocess.call('sudo -u debian jack_connect jack_tone:out JackTrip:send_2'.split())
    tone_proc.wait()
    wemust.dbg_print("Apparently jack_tone Exited or Died, with exit code %d" % tone_proc.returncode)


##################
# JACKTRIP funcs #
##################
# This has been modified in Python: always restarts jacktrip if it dies...
def jtrip_cli():
    while(1):
        jtbin_client_proc  = subprocess.Popen(('sudo -u debian' + ' ' +  JTBIN + ' ' + '-z -n2 -o' + ' ' + str(JT_OFFS) + ' ' + '-c' + ' ' + PEER_ADDRESS).split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT, bufsize=-1)
        jt_cli_dt = threading.Thread(target=wemust.dispatcherThread, args=(jtbin_client_proc.stdout, jt_cli_dbgs))
        jt_cli_dt.daemon = True # thread dies with the program
        jt_cli_dt.start()
        time.sleep(2)
        jtbin_client_proc.wait()
        wemust.dbg_print("Apparently jacktrip Exited or Died, with exit code %d" % jtbin_client_proc.returncode)


#
def conn_jt_Tx():
    subprocess.call('sudo -u debian jack_connect system:capture_1 JackTrip:send_1'.split())
    subprocess.call(('sudo -u debian jack_connect system:capture_' + str(NCHANNELS) + ' ' + 'JackTrip:send_2').split())

#
# This has been modified in Python: always restarts jacktrip if it dies...
def jtrip_srv():
    while(1):
        jtbin_server_proc = subprocess.Popen(('sudo -u debian' + ' ' + JTBIN + ' ' + '-z -n2 -s').split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT, bufsize=-1)
        jt_srv_dt = threading.Thread(target=wemust.dispatcherThread, args=(jtbin_server_proc.stdout, jt_srv_dbgs))
        jt_srv_dt.daemon = True # thread dies with the program
        jt_srv_dt.start()
        time.sleep(2)
        jtbin_server_proc.wait()
        wemust.dbg_print("Apparently jacktrip Exited or Died, with exit code %d" % jtbin_server_proc.returncode)


#
def conn_jt_Rx():
    subprocess.call('sudo -u debian jack_connect JackTrip:receive_1 system:playback_1'.split())
    subprocess.call(('sudo -u debian jack_connect JackTrip:receive_2 system:playback_' + str(NCHANNELS)).split())

def McBSP_settings():
    # DEFAULT McBSP2 AUDIO SETTINGS (if not already)
    subprocess.call('echo threshold | tee /sys/devices/68000000.ocp/49022000.mcbsp/dma_op_mode'.split())
    subprocess.call('echo 320 | tee /sys/devices/68000000.ocp/49022000.mcbsp/max_rx_thres'.split())
    subprocess.call('echo 320 | tee /sys/devices/68000000.ocp/49022000.mcbsp/max_tx_thres'.split())




############################
#### MAIN ! MAIN! MAIN! ####
############################
wemust.wInit()

if sys.argv[1] == '':
    case = 'start'
else:
    case = sys.argv[1]

if case == 'start':
    if PEER_ADDRESS == '':
        wemust.dbg_print("CRITICAL: PEER_ADDRESS is not set in wemust.conf - ABORTING")
    else:
        tune_max()
        McBSP_settings()
        jack_default()
        #conn_loopback()
        jtcli_thread = threading.Thread(target=jtrip_cli) #, args=()
        #jtcli_thread = threading.Thread(target=jtrip_srv) #, args=()
        jtcli_thread.start()
        time.sleep(2)
        #disc_loopback()
        #test_tone()
        conn_jt_Tx()
        conn_jt_Rx()
        jtcli_thread.join() # wait here for jtcli to join, it should never happen
        wemust.dbg_print("jackaudio.py DIED")
        # daemons will die now, unsafe for concerts: jackaudio could die, but jackd and jacktrip should keep alive...

elif case == 'stop':
    subprocess.Popen(["killall", JTBIN])
    subprocess.Popen(["killall", "jackd"])
    tune_avg()
    # kill any other instance of jackaudio
    for proc in psutil.process_iter():
        if any("jackaudio.py" in c for c in proc.cmdline):
            wemust.dbg_print("My PID: " + str(os.getpid()) )
            # useful debug print
            # wemust.dbg_print("candidate: " + str(proc.pid) )
            if proc.pid != os.getpid(): # do not suicide!
                # useful debug print
                wemust.dbg_print("Killing: " + str(proc.pid) )
                proc.kill()


elif case == 'cpumax':
    tune_max()

elif case == 'cpumin':
    tune_min()

elif case == 'cpuavg':
    tune_avg()

elif case == 'lbk_on':
    conn_loopback()

elif case == 'lbk_off':
    disc_loopback()

elif case == 'status':
    stat = subprocess.check_output(['ps', 'aux', '|', 'grep', '[j]ackd'], stdin=None, stderr=subprocess.STDOUT)
    wemust.dbg_print(stat)
    print(stat)

else:
    print("Usage: jackaudio {start|stop|cpumax|cpumin|cpuavg|lbk_on|lbk_off}")


