#! /usr/bin/python2
# wemust daemon: receives commands and sends info to wemust-netdbg
#
# Todo:
# - in the future I could think of only logging to file in jackaudio.py and doing the read and send here, so in case the former crashes I can still send everything that was logged
# - keyboard interrupt (ctrl-c) doesn't stop anything, the main keeps sleeping in .join()
# - added temp sensing, but need to pass a Ctrl+c to the get_temp child when quit_flag is raised
#
#
# author "Leonardo Gabrielli <l.gabrielli@univpm.it>


import socket
import subprocess
import time
import threading
import thread
import struct
import sys
import string

#######################
##### VARS - DEFS #####
#######################
import wemust
from wemust import USE_NETDBG, SERVICETYPE, UUID_LEN, NETDBG_MCAST_GROUP, NETDBG_BASE_PORT, NETDBG_PRINT_PORT, NETDBG_JACKD_PORT, \
                   NETDBG_JTSER_PORT, NETDBG_JTCLI_PORT, NETDBG_CTRL_PORT
threads = []
quitTimeout = 5 # seconds
global dbgSock


######################
## CONTROL   THREAD ##
######################

class ctrl_thread(threading.Thread):

    def __init__(self, quitFlag):
        threading.Thread.__init__(self, name="Control Thread")
        self.quitFlag = quitFlag

    def run_udp(self): # multicast UDP version, discarded
        print "[Wemust-daemon] Starting: " + self.name
        # Get lock to synchronize threads
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, proto=socket.IPPROTO_UDP)
        try:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        except AttributeError:
            pass # Some systems don't support SO_REUSEPORT
        sock.bind((NETDBG_MCAST_GROUP, NETDBG_CTRL_PORT)) # Allows binding to any interface. Security issues?
        mreq = struct.pack("=4sl", socket.inet_aton(NETDBG_MCAST_GROUP), socket.INADDR_ANY)
        try:
            sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        except:
            print("Warning: exception: " + str(sys.exc_info()[1]) + ". Connection will not be effective") # Todo: make it a popup dialog
            pass
        sock.settimeout(1) # if the sock is not totally blocking the thread can join

        while( not self.quitFlag.is_set() ):
            cmd = ''
            try:
                cmd, addr = sock.recvfrom(wemust.MAX_CTRL_CMD_LEN)
            except socket.timeout:
                pass
            # TODO check addr to avoid scums, a feasible strategy: register the first address received and then warn in case of other addrs
            if cmd != '':
                try:
                    ret = subprocess.check_output(cmd, shell=True, stdin=None, stderr=subprocess.STDOUT) # very dangerous, I know, to use shell=True
                    wemust.dbg_print(ret)
                except subprocess.CalledProcessError:
                    wemust.dbg_print("inconsistent shell command")
                    pass

    def run(self): # TCP version, current
        print "[Wemust-daemon] Starting: " + self.name
        # Get lock to synchronize threads
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        sock.bind(('', NETDBG_CTRL_PORT)) # Allows binding to any interface. Security issues?
        sock.settimeout(1) # if the sock is not totally blocking the thread can join
        # TODO check addr to avoid scums

        while( not self.quitFlag.is_set() ):
            conn = ''
            cmd = ''
            while conn == '': # socket methods are nonblocking: loop until conn is not empty
                try:
                    sock.listen(1) # Allows only one master to send commands
                    conn, addr = sock.accept()
                except socket.timeout:
                    pass
            # now we have a connection
            try:
                cmd = conn.recv(wemust.MAX_CTRL_CMD_LEN)
            except socket.timeout:
                pass
            if cmd != '':
                try:
                    proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, bufsize=-1) # very dangerous, I know
                    procDisp = threading.Thread(target=wemust.dispatcherThread, args=(proc.stdout, dbgSock)) # I could put sock.accept in the main and spanw a thread for each connection, but then I would use subprocess.check_output which does not send any data until the child is finished
                    threads.append(procDisp)
                    procDisp.daemon = True # thread dies with the program
                    procDisp.start()
                    # return
                except subprocess.CalledProcessError:
                    wemust.dbg_print("inconsistent shell command")
                    pass
            # the interaction cycle is ended, close connection and wait for new requests
            conn.send("Interaction ending")
            conn.close()


######################
## TEMPERATURE SENS ##
######################

class getTempThread(threading.Thread):
    def __init__(self, quitFlag):
        threading.Thread.__init__(self, name="Temperature Thread")
        self.quitFlag = quitFlag

    def run(self):
        try:
            print "[Wemust-daemon] Starting: " + self.name
            gTemp = subprocess.Popen('get_temp -m'.split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT, bufsize=-1)
            gTempDisp = threading.Thread(target=wemust.dispatcherThread, args=(gTemp.stdout, dbgSock))
            threads.append(gTempDisp)
            gTempDisp.daemon = True # thread dies with the program
            gTempDisp.start()
            gTemp.wait()
            gTempDisp.join()
        except (KeyboardInterrupt, SystemExit):
            tKill = subprocess.Popen('killall get_temp'.split())

        # must think of a way to send an interrupt to the child thread...
        #while( not self.quitFlag.is_set() ):
        #    gTempDisp.join(timeout=10) # check every timeout seconds, if the thread as joined


######################
## SABy LISTEN THR  ##
######################

class SAByListenThread(threading.Thread):

    def __init__(self, quitFlag, SyP):
        threading.Thread.__init__(self, name="SABy Listen Thread")
        self.quitFlag = quitFlag
        self.SyP = SyP

    def run(self): # does not follow netfind_listen() implementation: I am interested only in replying to SRC messages, not to ADV
        print "[Wemust-daemon] Starting: " + self.name
        # Get lock to synchronize threads
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, proto=socket.IPPROTO_UDP)
        try:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        except AttributeError:
            pass # Some systems don't support SO_REUSEPORT
        sock.bind((wemust.SABy_MCAST_GROUP, wemust.SABy_MCAST_PORT)) # Allows binding to any interface. Security issues?
        mreq = struct.pack("=4sl", socket.inet_aton(wemust.SABy_MCAST_GROUP), socket.INADDR_ANY)
        try:
            sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        except:
            print("Warning: exception: " + repr( sys.exc_info()[1] ) + ". Connection will not be effective")
            pass
        sock.settimeout(5) # with the sock not totally blocking the thread can join (with a little delay)

        while( not self.quitFlag.is_set() ):
            try:
                msg, addr = sock.recvfrom(wemust.SABy_MAX_MESSAGE_LEN)
                if msg:
                    #wemust.dbg_print(repr(msg))
                    if( string.find(msg, wemust.SABy_STRING) >= 0 ): # valid SABy message
                        if( string.find(msg, wemust.SABy_SEARCH_STRING) >= 0 ): # SRC message
                            spl = msg.split()
                            for line in spl:
                                if( string.find(line, wemust.SABy_TAG_STRING) >= 0 ):
                                    tags = line[len(wemust.SABy_TAG_STRING):len(line)]
                                    if (tags):
                                        for item in tags:             # look for a tag match
                                            if item in self.SyP.tag: # look for a tag match
                                                SyP.reply(addr)
                                                break
                                    else:
                                        SyP.reply(addr)
                        elif( string.find(msg, wemust.SABy_ANNOUNCE_STRING) >= 0 ):
                            pippo = 1
                            # do nothing, in future we may be interested in the perfect P2P of the original AES2013 work, not now
                        elif( string.find(msg, wemust.SABy_REPLY_STRING) >= 0 ):
                            pippo = 1
                            # do nothing, replying to the multicast is not accepted!
            except socket.timeout:
                pass


#######################
#####    MAIN     #####
#######################
dbgSock = wemust.wInit()
quitFlag = threading.Event()
SyP = wemust.SAByParams()

try:
    SyP.setFromFile(wemust.DEFAULT_WEMUSTRC_PATH)
    SyP.announce()
    sl_t = SAByListenThread(quitFlag, SyP)
    threads.append(sl_t)
    sl_t.start()
    ctrl_t = ctrl_thread(quitFlag)
    threads.append(ctrl_t)
    ctrl_t.start()
    temp_t = getTempThread(quitFlag)
    threads.append(temp_t)
    temp_t.start()
    #ctrl_t.join()
    while True:
        time.sleep(10)
except (KeyboardInterrupt, SystemExit):
    quitFlag.set()
    for t in range(len(threads)):
        ts = time.time()
        threads[t].join(quitTimeout)
        # if the time spent on join() is close to the timeout time it means join spent his time waiting
        # we want to track this down. We can't rely on .isAlive because it always returns true even if join
        # finished fine before the timeout
        if ( time.time() - ts > quitTimeout - 2):
            print("Warning on exit: thread still alive") # maybe it is locked on a resource
            print(threading.enumerate())